import typing as T
import asyncio
import contextlib
import psutil
import socket


class ExecutorError(Exception):
    pass


class CheckError(ExecutorError):
    pass


class TypeCheckError(CheckError):
    pass


class RangeCheckError(CheckError):
    pass


class CheckAttrRange(object):
    valid_range: T.Iterable[T.Any] = []
    attr = "__"

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.attr)

    def check(self, obj, value):
        if value not in self.valid_range:
            raise RangeCheckError(
                f"{obj}'s {type(self)} attr should be "
                f"one of: {self.valid_range}"
            )

    def __set__(self, obj, value):
        self.check(obj, value)
        setattr(obj, self.attr, value)


Checker = T.Callable[[T.Any], bool]


class CheckAttrType(object):
    valid_type: T.List[T.Union[Checker, type]] = []
    attr = "__"

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.attr)

    def __set__(self, obj, value):
        check_passed = []
        for tp in self.valid_type:
            if isinstance(tp, type):
                passed = isinstance(value, tp)
            else:
                assert isinstance(tp, T.Callable)
                passed = tp(value)
            check_passed.append(passed)
        is_valid_type = any(check_passed)
        if not is_valid_type:
            raise TypeCheckError(
                f"{obj}'s {type(self)} attr should in type: "
                f"{self.valid_type}"
            )
        setattr(obj, self.attr, value)


@contextlib.contextmanager
def event_loop():
    loop, is_new_loop = get_event_loop()
    try:
        yield loop
    finally:
        if is_new_loop:
            loop.close()


def get_event_loop() -> T.Tuple[asyncio.AbstractEventLoop, bool]:
    is_new_loop = False
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        is_new_loop = True
    return loop, is_new_loop


class PortManager():

    used_port: T.Set[int] = set()

    @classmethod
    def get_port(cls) -> int:
        while True:
            port = cls.find_free_port()
            if port in cls.used_port:  # pragma: no cover
                continue
            else:
                cls.consume_port(port)
                return port

    @classmethod
    def consume_port(cls, port: int):
        cls.used_port.add(port)

    @classmethod
    def release_port(cls, port: int):
        cls.used_port.remove(port)

    @staticmethod
    def find_free_port() -> int:
        """https://stackoverflow.com/a/45690594/8500469"""
        with contextlib.closing(
                socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('', 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            port = s.getsockname()[1]
        return port

    @staticmethod
    def process_has_port(target_pid: int, ip: str, port: int) -> bool:
        p = psutil.Process(target_pid)
        addrs = [
            (c.laddr.ip, c.laddr.port) for c in p.connections()
        ]
        return (ip, port) in addrs


def get_callable_name(callable) -> str:
    if hasattr(callable, "func"):
        inner_func = getattr(callable, "func")
        return get_callable_name(inner_func)
    elif hasattr(callable, "__name__"):
        return getattr(callable, "__name__")
    elif hasattr(callable, "__class__"):
        return getattr(callable, "__class__").__name__
    else:  # pragma: no cover
        return str(callable)



import typing as T
from pathlib import Path

from diskcache import Cache

from .job.base import Job
from .job.utils import (
    valid_job_statuses, JobStatusType, ExecutorError
)

if T.TYPE_CHECKING:
    from .core import Engine


class JobNotFoundError(ExecutorError):
    """Job not found error."""
    def __init__(self, job_id: str):
        self.job_id = job_id
        super().__init__(f"Job {job_id} not found.")


class JobStore():
    """Store jobs.

    Attributes:
        cache: The cache on disk(Use diskcache package).
            See:
            [disk-cache's doc](https://github.com/grantjenks/python-diskcache)
            for more details.
        mem: In memory store.
    """

    cache: T.Optional[Cache]
    mem: T.Dict[str, Job]

    def __init__(self, cache_path: T.Optional[Path] = None):
        """Init.

        Args:
            cache_path: Cache path.
        """
        if cache_path is not None:
            self.cache = Cache(str(cache_path))
        else:
            self.cache = None
        self.mem: T.Dict[str, Job] = dict()

    @classmethod
    def load_from_cache(cls, path: Path):
        """Load from cache."""
        store = cls(path)
        store.update_from_cache()
        return store

    def update_from_cache(self, clear_old=False):
        """Update from cache."""
        if clear_old:
            self.mem.clear()
        if self.cache is not None:
            for key in self.cache:
                job = self.get_from_cache(key)
                self.mem[key] = job

    def get_from_cache(self, key: str) -> Job:
        """Get from cache."""
        if self.cache is None:
            raise RuntimeError("No cache")
        bytes_ = self.cache[key]
        job = Job.deserialization(bytes_)
        return job

    def set_to_cache(self, key: str, val: Job):
        """Set job to cache."""
        bytes_ = val.serialization()
        if self.cache is not None:
            self.cache[key] = bytes_

    def __setitem__(self, key: str, val: Job):
        self.mem[key] = val
        if self.cache is not None:
            self.set_to_cache(key, val)

    def __getitem__(self, key: str) -> Job:
        return self.mem[key]

    def __contains__(self, key: str) -> bool:
        return key in self.mem

    def clear(self):
        """Clear all jobs."""
        self.mem.clear()
        if self.cache is not None:
            self.cache.clear()

    def pop(self, key: str) -> Job:
        """Pop a job from store."""
        job = self.mem.pop(key)
        if self.cache is not None:
            self.cache.pop(key)
        return job

    def values(self) -> T.List[Job]:
        """Get all values(Job)."""
        vals = list(self.mem.values())
        return vals

    def keys(self) -> T.List[str]:
        """Get all keys(Job's id)."""
        return list(self.mem.keys())

    def items(self) -> T.List[T.Tuple[str, Job]]:
        """Get all key-value pairs."""
        return list(self.mem.items())

    def __del__(self):
        if self.cache is not None:
            self.cache.close()

    def __len__(self):
        return len(self.mem)


class Jobs:
    """Jobs manager.

    Attributes:
        pending: Pending jobs.
        running: Running jobs.
        done: Done jobs.
        failed: Failed jobs.
        cancelled: Cancelled jobs.
    """
    valid_statuses = valid_job_statuses
    pending: JobStore
    running: JobStore
    done: JobStore
    failed: JobStore
    cancelled: JobStore

    def __init__(self, cache_path: Path):
        self.cache_path = cache_path
        self._stores: T.Dict[str, JobStore] = {}
        s: str
        for s in self.valid_statuses:
            path = cache_path / s
            if path.exists():
                store = JobStore.load_from_cache(path)
            else:
                store = JobStore(path)
            self._stores[s] = store
        self.set_attrs_for_read()

    def update_from_cache(self, clear_old=True):
        """Update jobs from cache."""
        for store in self._stores.values():
            store.update_from_cache(clear_old=clear_old)

    def set_attrs_for_read(self):
        self.pending = self._stores['pending']
        self.running = self._stores['running']
        self.done = self._stores['done']
        self.failed = self._stores['failed']
        self.cancelled = self._stores['cancelled']

    def set_engine(self, engine: "Engine"):
        """Set engine for all jobs."""
        for job in self.all_jobs():
            job.engine = engine

    def clear(self, statuses: T.List[JobStatusType]):
        """Clear jobs by status."""
        for s in statuses:
            self._stores[s].clear()

    def clear_non_active(self):
        """Clear non-active jobs."""
        self.clear(["done", "failed", "cancelled"])

    def clear_all(self):
        """Clear all jobs."""
        self.clear(self.valid_statuses)

    def add(self, job: Job):
        """Add job to store."""
        store = self._stores[job.status]
        store[job.id] = job

    def remove(self, job: Job):
        """Remove job from store."""
        for tp in self.valid_statuses:
            store = self._stores[tp]
            if job.id in store:
                store.pop(job.id)

    def move_job_store(
            self, job: Job,
            new_status: JobStatusType,
            old_status: T.Optional[JobStatusType] = None):
        """Move job to another store."""
        if old_status is None:
            old_status = job.status
        if old_status == new_status:
            return
        old_store = self._stores[old_status]
        new_store = self._stores[new_status]
        new_store[job.id] = old_store.pop(job.id)

    def get_job_by_id(self, job_id: str) -> Job:
        """Get job by id."""
        for status in self.valid_statuses:
            store = self._stores[status]
            if job_id in store:
                return store[job_id]
        raise JobNotFoundError(job_id)

    def __contains__(self, job: T.Union[str, Job]):
        if isinstance(job, Job):
            job_id = job.id
        else:
            job_id = job
        try:
            self.get_job_by_id(job_id)
            return True
        except JobNotFoundError:
            return False

    def all_jobs(self) -> T.List[Job]:
        """Get all jobs."""
        return list(iter(self))

    def __iter__(self):
        """Iterate all jobs."""
        for status in self.valid_statuses:
            store = self._stores[status]
            for job in store.values():
                yield job

    def __len__(self):
        return sum(len(store) for store in self._stores.values())



from .core import Engine, EngineSetting
from .job import LocalJob, ThreadJob, ProcessJob

__version__ = '0.1.3'

__all__ = [
    'Engine', 'EngineSetting',
    'LocalJob', 'ThreadJob', 'ProcessJob'
]



import copy
import shlex
import typing as T
import subprocess as subp
from pathlib import Path

from ..condition import Condition
from ..thread import ThreadJob
from ..base import Job

from cmd2func.runner import ProcessRunner


def SubprocessJob(
    cmd: str, record_cmd: bool = True,
    base_class: T.Type[Job] = ThreadJob,
    callback: T.Optional[T.Callable[[T.Any], None]] = None,
    error_callback: T.Optional[T.Callable[[Exception], None]] = None,
    retries: int = 0,
    retry_time_delta: float = 0.0,
    name: T.Optional[str] = None,
    condition: T.Optional[Condition] = None,
    wait_time_delta: float = 0.01,
    redirect_out_err: bool = False,
    target_dir: str = "$current_dir",
    **attrs
):
    """Create a job that runs a subprocess.

    Args:
        cmd: The command to run.
        record_cmd: Whether to record the command to a file.
        base_class: The base class of the job.
        callback: The callback function.
        error_callback: The error callback function.
        retries: The number of retries.
        retry_time_delta: The time delta between retries.
        name: The name of the job.
        condition: The condition of the job.
        wait_time_delta: The time delta between each check.
        redirect_out_err: Whether to redirect stdout and stderr to files.
        target_dir: The target directory path for run the command.
            Use '$cache_dir' to represent the cache directory of the job.
            Use '$current_dir' to represent the current directory of the job.
            Default is '$current_dir'.
        **attrs: Other attributes of the job.
    """
    class _SubprocessJob(base_class):  # type: ignore
        cmd: str
        target_dir: str

        repr_attrs = [
            ('status', lambda self: self.status),
            ('id', lambda self: self.id),
            ('cmd', lambda self: self.cmd),
            ('target_dir', lambda self: self.target_dir),
            ('base_class', lambda _: base_class.__name__),
            ("condition", lambda self: self.condition),
            ("retry_remain", lambda self: self.retry_remain),
        ]

        def __init__(self) -> None:
            self.cmd = cmd
            self.record_cmd = record_cmd
            nonlocal name
            if name is None:
                name = cmd.split()[0]
            self.target_dir = target_dir
            attrs.update({
                'cmd': cmd,
                'target_dir': target_dir,
            })
            super().__init__(
                lambda x: x,
                callback=callback,
                error_callback=error_callback,
                retries=retries,
                retry_time_delta=retry_time_delta,
                name=name,
                condition=condition,
                wait_time_delta=wait_time_delta,
                redirect_out_err=redirect_out_err,
                **attrs
            )

        def resolve_target_dir(self, target_dir: str) -> str:
            if target_dir == "$cache_dir":
                return self.cache_dir.resolve().as_posix()
            elif target_dir == "$current_dir":
                return Path.cwd().resolve().as_posix()
            else:
                return Path(target_dir).resolve().as_posix()

        def process_func(self):
            cmd = copy.copy(self.cmd)
            record_cmd = copy.copy(self.record_cmd)
            target_dir = self.resolve_target_dir(self.target_dir)
            self.attrs.update({
                'target_dir': target_dir,
            })
            self.target_dir = target_dir

            cache_dir = self.cache_dir.resolve()
            path_sh = cache_dir / 'command.sh'

            def record_command():
                with open(path_sh, 'w') as f:
                    f.write(cmd + "\n")

            if self.redirect_out_err:
                path_stdout = cache_dir / 'stdout.txt'
                path_stderr = cache_dir / 'stderr.txt'

                def run_cmd():  # pragma: no cover
                    runner = ProcessRunner(cmd)
                    runner.run(cwd=target_dir)
                    with open(path_stdout, 'w') as fo, \
                         open(path_stderr, 'w') as fe:
                        retcode = runner.write_stream_until_stop(fo, fe)
                    return retcode
            else:
                def run_cmd():
                    p = subp.Popen(shlex.split(cmd), cwd=target_dir)
                    retcode = p.wait()
                    return retcode

            def func():
                if record_cmd:
                    record_command()
                retcode = run_cmd()
                if retcode > 0:
                    raise subp.SubprocessError(
                        f"Command '{cmd}' run failed, return code: {retcode}")
                return retcode
            self.func = func

    return _SubprocessJob()



import typing as T
from dataclasses import dataclass
from pathlib import Path
import asyncio
from threading import Thread
import time
import concurrent.futures

from .base import ExecutorObj
from .job.base import Job
from .manager import Jobs
from .log import logger

if T.TYPE_CHECKING:
    from dask.distributed import Client


@dataclass
class EngineSetting:
    """Engine setting.

    Args:
        max_thread_jobs: Maximum number of thread jobs,
            if not set, no limit.
        max_process_jobs: Maximum number of process jobs,
            if not set, no limit.
        max_dask_jobs: Maximum number of dask jobs,
            if not set, no limit.
        max_jobs: Maximum number of jobs,
            if not set, no limit.
        cache_path: Cache path,
            if not set, will create a cache directory in
            .executor/{engine.id}.
        print_traceback: Whether to print traceback when job failed.
    """
    max_thread_jobs: T.Optional[int] = None
    max_process_jobs: T.Optional[int] = None
    max_dask_jobs: T.Optional[int] = None
    max_jobs: T.Optional[int] = 20
    cache_path: T.Optional[str] = None
    print_traceback: bool = True


@dataclass
class Resource:
    """Resource of engine.

    Args:
        n_thread: Number of thread jobs.
        n_process: Number of process jobs.
        n_dask: Number of dask jobs.
        n_job: Number of jobs.
    """
    n_thread: T.Union[int, float]
    n_process: T.Union[int, float]
    n_dask: T.Union[int, float]
    n_job: T.Union[int, float]


class Engine(ExecutorObj):
    def __init__(
            self,
            setting: T.Optional[EngineSetting] = None,
            jobs: T.Optional[Jobs] = None,
            loop: T.Optional[asyncio.AbstractEventLoop] = None,
            ) -> None:
        """
        Args:
            setting: Engine setting. Defaults to None.
            jobs: Jobs manager. Defaults to None.
            loop: Event loop. Defaults to None.

        Attributes:
            setting: Engine setting.
            resource: Resource of engine.
            jobs: Jobs manager.
        """
        super().__init__()
        if setting is None:
            setting = EngineSetting()
        self.setting = setting
        self.print_traceback = False
        self.setup_by_setting()
        if jobs is None:
            jobs = Jobs(self.cache_dir / "jobs")
        self.jobs: Jobs = jobs
        self._dask_client: T.Optional["Client"] = None
        self._loop = loop
        self._loop_thread: T.Optional[Thread] = None

    def __repr__(self) -> str:
        return f"<Engine id={self.id}>"

    def __str__(self) -> str:
        return repr(self)

    @property
    def loop(self):
        """Event loop of engine."""
        if self._loop is None:
            loop = asyncio.new_event_loop()
            logger.info(f"{self} created new event loop.")
            self._loop = loop
        return self._loop

    @loop.setter
    def loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop

    def start(self):
        """Start event loop thread."""
        def run_in_thread(loop: asyncio.AbstractEventLoop):
            logger.info(f"{self} start event loop.")
            asyncio.set_event_loop(loop)
            loop.run_forever()

        if self._loop_thread is None:
            self._loop_thread = Thread(
                target=run_in_thread,
                args=(self.loop,), daemon=True
            )

        if not self._loop_thread.is_alive():
            logger.info(f"{self} start event loop thread.")
            self._loop_thread.start()
        else:
            logger.warning(f"Event loop thread of {self} is already running.")

    def stop(self):
        """Stop event loop thread."""
        loop = self._loop
        if (loop is None) or (self._loop_thread is None):
            logger.warning(f"Event loop thread of {self} is not running.")
            return

        if not self._loop_thread.is_alive():
            logger.warning(f"Event loop thread of {self} is already closed.")
        else:
            logger.info(f"{self} stop event loop.")
            loop.call_soon_threadsafe(loop.stop)
            self._loop_thread.join()
            if self._dask_client is not None:
                asyncio.run_coroutine_threadsafe(
                    self._dask_client.close(), loop)

    def __enter__(self):
        self.start()
        from . launcher.core import set_default_engine
        set_default_engine(self)
        return self

    def __exit__(self, *args):
        self.stop()
        from . launcher.core import set_default_engine
        set_default_engine(None)

    def setup_by_setting(self):
        setting = self.setting
        logger.info(f"Load setting: {setting}")
        self.resource = Resource(
            n_thread=setting.max_thread_jobs or float('inf'),
            n_process=setting.max_process_jobs or float('inf'),
            n_dask=setting.max_dask_jobs or float('inf'),
            n_job=setting.max_jobs or float('inf'),
        )
        self.cache_dir = self.get_cache_dir()
        self.print_traceback = setting.print_traceback

    def submit(self, *jobs: Job):
        """Submit job to engine"""
        if (self._loop_thread is None) or\
           (not self._loop_thread.is_alive()):
            raise RuntimeError(
                f"Event loop thread of {self} is not running."
                " Please use engine as context manager or call engine.start()."
            )
        fut = asyncio.run_coroutine_threadsafe(
            self.submit_async(*jobs), self.loop)
        fut.result()

    async def submit_async(self, *jobs: Job):
        """Asynchronous interface for submit jobs to engine."""
        for job in jobs:
            if job.status == "created":
                job.engine = self
                job._status = "pending"
                self.jobs.add(job)
            else:
                job.status = "pending"
            assert job.engine is self, "Job engine is not this engine."
            await job.emit()

    def remove(self, job: Job):
        """Remove job from engine."""
        if job.status in ('running', 'pending'):
            fut = asyncio.run_coroutine_threadsafe(
                job.cancel(), self.loop)
            fut.result()
        logger.info(f"Remove job from engine: {job}")
        self.jobs.remove(job)

    def cancel(self, *jobs: Job):
        """Cancel a job."""
        futures = []
        for job in jobs:
            fut = asyncio.run_coroutine_threadsafe(
                job.cancel(), self.loop)
            futures.append(fut)
        concurrent.futures.wait(futures)

    async def cancel_all_async(self):
        """Cancel all pending and running jobs."""
        running = self.jobs.running.values()
        pending = self.jobs.pending.values()
        tasks = []
        for job in (pending + running):
            tasks.append(job.cancel())
        await asyncio.gather(*tasks)

    def cancel_all(self):
        """Cancel all pending and running jobs."""
        fut = asyncio.run_coroutine_threadsafe(
            self.cancel_all_async(), self.loop)
        fut.result()

    def wait_job(
            self, job: Job,
            timeout: T.Optional[float] = None,
            ) -> T.Optional[T.Any]:
        """Block until job is finished or timeout.
        Return job result if job is done.

        Args:
            job: Job to wait.
            timeout: Timeout in seconds.
        """
        fut = asyncio.run_coroutine_threadsafe(
            job.join(timeout=timeout), self.loop)
        fut.result()
        if job.status == "done":
            return job.result()
        else:
            return None

    def wait(
            self,
            timeout: T.Optional[float] = None,
            time_delta: float = 0.2):
        """Block until all jobs are finished or timeout.

        Args:
            timeout: Timeout in seconds.
            time_delta: Time interval to check job status.
        """
        total_time = timeout if timeout is not None else float('inf')
        while True:
            n_wait_jobs = len(self.jobs.running) + len(self.jobs.pending)
            if n_wait_jobs == 0:
                break
            if total_time <= 0:
                break
            time.sleep(time_delta)
            total_time -= time_delta

    async def join(self, timeout: T.Optional[float] = None):
        """Asynchronous interface for wait all jobs."""
        running = self.jobs.running.values()
        pending = self.jobs.pending.values()
        tasks = [
            asyncio.create_task(job.join())
            for job in (running + pending)
        ]
        await asyncio.wait(tasks, timeout=timeout)

    def get_cache_dir(self) -> Path:
        """Get cache directory for engine."""
        cache_path = self.setting.cache_path
        if cache_path is not None:
            path = cache_path
        else:
            path = f".executor/{self.id}"
        path_obj = Path(path)
        path_obj.mkdir(parents=True, exist_ok=True)
        return path_obj

    @property
    def dask_client(self):
        from .job.dask import get_default_client
        if self._dask_client is None:
            self._dask_client = get_default_client()
        return self._dask_client

    @dask_client.setter
    def dask_client(self, client: "Client"):
        if not client.asynchronous:
            raise ValueError("Dask client must be asynchronous.")
        self._dask_client = client



import time
import typing as T
import copy
import sys

from loky.backend.process import LokyProcess
from cmd2func.runner import ProcessRunner

from ..process import ProcessJob
from ..base import Job
from ..condition import Condition
from ...utils import PortManager, ExecutorError


LauncherFunc = T.Callable[[str, int], None]


def WebappJob(
    web_launcher: T.Union[LauncherFunc, str],
    ip: str = "127.0.0.1", port: T.Optional[int] = None,
    base_class: T.Type[Job] = ProcessJob,
    check_times: int = 6,
    check_delta: float = 1.0,
    callback: T.Optional[T.Callable[[T.Any], None]] = None,
    error_callback: T.Optional[T.Callable[[Exception], None]] = None,
    retries: int = 0,
    retry_time_delta: float = 0.0,
    name: T.Optional[str] = None,
    condition: T.Optional[Condition] = None,
    wait_time_delta: float = 0.01,
    redirect_out_err: bool = False,
    **attrs
):
    """Create a job that runs a web app.

    Args:
        web_launcher: The function to launch the web app.
            The function should accept two arguments: ip and port.
        ip: The ip address of the web app.
        port: The port of the web app. If None, will find a free port.
        base_class: The base class of the job.
        check_times: The number of times to check the web app.
        check_delta: The time delta between each check.
        callback: The callback function.
        error_callback: The error callback function.
        retries: The number of retries.
        retry_time_delta: The time delta between retries.
        name: The name of the job.
        condition: The condition of the job.
        wait_time_delta: The time delta between each check.
        redirect_out_err: Whether to redirect stdout and stderr to files.
        **attrs: Other attributes of the job.
    """
    class _WebappJob(base_class):  # type: ignore
        ip: str
        port: T.Optional[int]

        repr_attrs = [
            ('status', lambda self: self.status),
            ('id', lambda self: self.id),
            ('addr', lambda self: f'{self.ip}:{self.port}'),
            ('base_class', lambda _: base_class.__name__),
            ("condition", lambda self: self.condition),
            ("retry_remain", lambda self: self.retry_remain),
        ]

        def __init__(self) -> None:
            self.ip = ip
            if ip not in ("127.0.0.1", "localhost", "0.0.0.0"):
                raise NotImplementedError(
                    "WebappJob now only support launch in local mechine.")
            self.port = port
            self.check_web_launcher(web_launcher)
            self.web_launcher = web_launcher
            self.check_times = check_times
            self.check_delta = check_delta
            if isinstance(web_launcher, str):
                attrs.update({"cmd": web_launcher})
            super().__init__(
                lambda x: x, callback=callback,
                error_callback=error_callback,
                retries=retries,
                retry_time_delta=retry_time_delta,
                name=name,
                condition=condition,
                wait_time_delta=wait_time_delta,
                redirect_out_err=redirect_out_err,
                **attrs
            )

        @staticmethod
        def check_web_launcher(web_launcher: T.Union[LauncherFunc, str]):
            if isinstance(web_launcher, str):
                cmd_temp = web_launcher
                if ('{ip}' not in cmd_temp) or ('{port}' not in cmd_temp):
                    raise ValueError(
                        "web_launcher should has ip and port placeholder.")
            else:
                if not callable(web_launcher):
                    raise TypeError(
                        "web_launcher should be a callable object or str.")

        def consume_resource(self) -> bool:
            if super().consume_resource():
                if self.port is None:
                    self.port = PortManager.get_port()
                    self.attrs.update({"address": f"{self.ip}:{self.port}"})
                else:
                    PortManager.consume_port(self.port)
                return True
            else:  # pragma: no cover
                return False

        def release_resource(self) -> bool:
            if self.port is None:  # pragma: no cover
                return False
            if super().release_resource():
                PortManager.release_port(self.port)
                return True
            else:  # pragma: no cover
                return False

        def process_func(self):
            web_launcher = copy.copy(self.web_launcher)
            if self.port is None:  # pragma: no cover
                raise ExecutorError("Unreachable code.")
            ip, port = copy.copy(self.ip), copy.copy(self.port)
            check_times = copy.copy(self.check_times)
            check_delta = copy.copy(self.check_delta)

            def check_port(pid: int) -> bool:  # pragma: no cover
                for _ in range(check_times):
                    time.sleep(check_delta)
                    if PortManager.process_has_port(pid, ip, port):
                        return True
                    print(f"Process is not listen on {ip}:{port}. Try again.")
                return False

            if isinstance(web_launcher, str):
                cmd = web_launcher.format(ip=ip, port=port)

                def func():  # pragma: no cover
                    runner = ProcessRunner(cmd)
                    runner.run()
                    if check_port(runner.proc.pid):
                        retcode = runner.write_stream_until_stop(
                            sys.stdout, sys.stderr)
                        sys.exit(retcode)
                    else:
                        runner.proc.terminate()
                        raise IOError(f"Process is not listen on {ip}:{port}.")
            else:
                def func():  # pragma: no cover
                    proc = LokyProcess(target=web_launcher, args=(ip, port))
                    proc.start()
                    pid = proc.pid
                    if check_port(pid):
                        proc.join()
                    else:
                        proc.terminate()
                        raise IOError(f"Process is not listen on {ip}:{port}.")

            self.func = func
            super().process_func()
    return _WebappJob()



import typing as T
from datetime import datetime
from dataclasses import dataclass

from .utils import JobStatusType

if T.TYPE_CHECKING:
    from ..core import Engine


@dataclass
class Condition():
    """Base class for condition"""

    def satisfy(self, engine: "Engine") -> bool:  # pragma: no cover
        """Check if the condition is satisfied."""
        return True


@dataclass
class AfterAnother(Condition):
    """Condition that the job is executed after
    another job is done/failed/cancelled.

    Attributes:
        job_id: The id of the job.
        statuses: The statuses of the job that satisfy the condition.
    """

    job_id: str
    statuses: T.Iterable[JobStatusType] = ("done", "failed", "cancelled")

    def satisfy(self, engine):
        """Check if the condition is satisfied."""
        try:
            another = engine.jobs.get_job_by_id(self.job_id)
        except Exception:
            return False
        if another.status in self.statuses:
            return True
        else:
            return False


@dataclass
class AfterOthers(Condition):
    """Condition that the job is executed after
    other jobs are done/failed/cancelled.

    Attributes:
        job_ids: The ids of the jobs.
        statuses: The statuses of the jobs that satisfy the condition.
        mode: The mode of the condition. If it is 'all', the job is executed
            after all other jobs are done/failed/cancelled. If it is 'any',
            the job is executed after any other job is done/failed/cancelled.
    """

    job_ids: T.List[str]
    statuses: T.Iterable[JobStatusType] = ("done", "failed", "cancelled")
    mode: T.Literal['all', 'any'] = "all"

    def satisfy(self, engine):
        """Check if the condition is satisfied."""
        other_job_satisfy = []
        for id_ in self.job_ids:
            try:
                job = engine.jobs.get_job_by_id(id_)
            except Exception:
                other_job_satisfy.append(False)
                continue
            s_ = job.status in self.statuses
            other_job_satisfy.append(s_)
        if self.mode == 'all':
            return all(other_job_satisfy)
        else:
            return any(other_job_satisfy)


@dataclass
class AfterTimepoint(Condition):
    """Condition that the job is executed after a timepoint.

    Attributes:
        timepoint: The timepoint.
    """

    timepoint: datetime

    def satisfy(self, engine):
        """Check if the condition is satisfied."""
        if datetime.now() > self.timepoint:
            return True
        else:
            return False


@dataclass
class Combination(Condition):
    """Base class for combination of conditions.

    Attributes:
        conditions: The sub-conditions.
    """
    conditions: T.List[Condition]


@dataclass
class AllSatisfied(Combination):
    """Condition that the job is executed after all
    sub-conditions are satisfied.

    Attributes:
        conditions: The sub-conditions.
    """

    conditions: T.List[Condition]

    def satisfy(self, engine):
        """Check if the condition is satisfied."""
        return all([c.satisfy(engine) for c in self.conditions])


@dataclass
class AnySatisfied(Combination):
    """Condition that the job is executed after any
    sub-condition is satisfied.

    Attributes:
        conditions: The sub-conditions.
    """

    conditions: T.List[Condition]

    def satisfy(self, engine):
        """Check if the condition is satisfied."""
        return any([c.satisfy(engine) for c in self.conditions])



import os
import typing as T
from pathlib import Path
from functools import update_wrapper


class ChDir(object):
    def __init__(self, func: T.Callable, target_dir: Path):
        update_wrapper(func, self)
        self.func = func
        self.target_dir = target_dir

    def __call__(self, *args, **kwargs) -> T.Any:
        old = os.getcwd()
        os.chdir(self.target_dir)
        res = self.func(*args, **kwargs)
        os.chdir(old)
        return res



import typing as T
from datetime import datetime

from ..utils import CheckAttrRange, ExecutorError


if T.TYPE_CHECKING:
    from .base import Job


JobStatusType = T.Literal['pending', 'running', 'failed', 'done', 'cancelled']
valid_job_statuses = JobStatusType.__args__  # type: ignore


class JobStatusAttr(CheckAttrRange):
    valid_range: T.Iterable[JobStatusType] = valid_job_statuses
    attr = "_status"

    def __set__(self, obj: "Job", value: JobStatusType):
        self.check(obj, value)
        old_status = getattr(obj, self.attr)
        setattr(obj, self.attr, value)
        if obj.engine is not None:
            obj.engine.jobs.move_job_store(
                obj, value, old_status)
        if value in ('done', 'failed', 'cancelled'):
            obj.stoped_time = datetime.now()


class JobEmitError(ExecutorError):
    pass


class InvalidStateError(ExecutorError):
    pass



import typing as T
import inspect
import functools
from copy import copy

from funcdesc import parse_func
from cmd2func.core import Cmd2Func

from ..core import Engine
from ..job import Job, LocalJob, ThreadJob, ProcessJob
from ..job.extend import SubprocessJob, WebappJob
from ..utils import get_callable_name


Job_or_ExtJob = T.Union[T.Type[Job], T.Callable[..., Job]]


job_type_classes: T.Dict[str, Job_or_ExtJob] = {
    'local': LocalJob,
    'thread': ThreadJob,
    'process': ProcessJob,
    'subprocess': SubprocessJob,
    'webapp': WebappJob,
}


JOB_TYPES = T.Literal[
    'local', 'thread', 'process',
    'subprocess', 'webapp'
]


_engine: T.Optional[Engine] = None


def get_default_engine() -> Engine:
    """Get the default engine."""
    global _engine
    if _engine is None:
        _engine = Engine()
    return _engine


def set_default_engine(engine: Engine):
    """Set the default engine."""
    global _engine
    _engine = engine


class LauncherBase(object):
    def __init__(
            self, target_func: T.Union[T.Callable, Cmd2Func],
            engine: T.Optional['Engine'] = None,
            job_type: JOB_TYPES = 'process',
            name: T.Optional[str] = None,
            description: T.Optional[str] = None,
            tags: T.Optional[T.List[str]] = None,
            **job_attrs):
        self._engine = engine
        self.target_func = target_func
        self.__signature__ = inspect.signature(target_func)
        self.job_type = job_type
        if isinstance(target_func, Cmd2Func):
            if self.job_type != 'webapp':
                self.job_type = 'subprocess'
        self.desc = parse_func(target_func)
        self.name = name or get_callable_name(target_func)
        self.description = description or self.target_func.__doc__
        functools.update_wrapper(self, target_func)
        self.tags = tags or []
        self.job_attrs = job_attrs

    @property
    def engine(self) -> Engine:
        """Get the engine of the launcher."""
        if self._engine is None:
            self._engine = get_default_engine()
        return self._engine

    @engine.setter
    def engine(self, engine: Engine):
        """Set the engine of the launcher."""
        self._engine = engine

    def create_job(self, args: tuple, kwargs: dict, **attrs) -> 'Job':
        """Create a job from the launcher."""
        job_attrs = copy(self.job_attrs)
        job_attrs.update(attrs)
        job_class = job_type_classes[self.job_type]
        if isinstance(self.target_func, Cmd2Func):
            assert job_class in (SubprocessJob, WebappJob)
            cmd_or_gen = self.target_func.get_cmd_str(*args, **kwargs)
            if isinstance(cmd_or_gen, str):
                cmd = cmd_or_gen
            else:
                cmd = next(cmd_or_gen)
            job = job_class(cmd, **job_attrs)  # type: ignore
        else:
            if job_class is WebappJob:
                kwargs = copy(kwargs)
                kwargs.update(job_attrs)
                job = job_class(self.target_func, *args, **kwargs)
            else:
                job = job_class(
                    self.target_func, args, kwargs,
                    **job_attrs
                )
        return job

    @staticmethod
    def _fetch_result(job: 'Job') -> T.Any:
        if job.status == "failed":
            raise job.exception()
        elif job.status == "cancelled":
            raise RuntimeError("Job cancelled")
        else:
            return job.result()


class SyncLauncher(LauncherBase):

    @property
    def async_mode(self):
        """Check if the launcher is in async mode."""
        return False

    def submit(self, *args, **kwargs) -> Job:
        """Submit a job to the engine."""
        job = self.create_job(args, kwargs)
        self.engine.submit(job)
        return job

    def __call__(self, *args, **kwargs) -> T.Any:
        """Submit a job to the engine and wait for the result."""
        job = self.submit(*args, **kwargs)
        self.engine.wait_job(job)
        return self._fetch_result(job)

    def to_async(self) -> "AsyncLauncher":
        """Convert the launcher to async mode."""
        return AsyncLauncher(
            self.target_func, self._engine, self.job_type,
            self.name, self.description, self.tags, **self.job_attrs,
        )


class AsyncLauncher(LauncherBase):
    @property
    def async_mode(self):
        """Check if the launcher is in async mode."""
        return True

    async def submit(self, *args, **kwargs):
        """Submit a job to the engine."""
        job = self.create_job(args, kwargs)
        await self.engine.submit_async(job)
        return job

    async def __call__(self, *args, **kwargs) -> T.Any:
        """Submit a job to the engine and wait for the result."""
        job = await self.submit(*args, **kwargs)
        await job.join()
        return self._fetch_result(job)

    def to_sync(self) -> "SyncLauncher":
        """Convert the launcher to sync mode."""
        return SyncLauncher(
            self.target_func, self._engine, self.job_type,
            self.name, self.description, self.tags, **self.job_attrs,
        )


def launcher(
        func: T.Optional[T.Union[T.Callable, Cmd2Func]] = None,
        engine: T.Optional['Engine'] = None,
        async_mode: bool = False,
        job_type: JOB_TYPES = 'process',
        name: T.Optional[str] = None,
        description: T.Optional[str] = None,
        tags: T.Optional[T.List[str]] = None,
        **job_attrs: T.Dict):
    """Create a launcher for a function.

    Args:
        func: The function to be launched. If the function is instance of
            [Cmd2Func](https://github.com/Nanguage/cmd2func),
            the launcher will be in subprocess mode,
            will launch `SubprocessJob` on each submit.
        engine: The engine to use. If not specified, the default engine
            will be used.
        async_mode: If True, the launcher will be AsyncLauncher.
        job_type: The job type to use. Default is 'process'.
        name: The name of the launcher.
        description: The description of the launcher.
        tags: The tags of the launcher.
        job_attrs: The attributes for creating the job.
    """
    if func is None:
        return functools.partial(
            launcher, engine=engine, async_mode=async_mode,
            job_type=job_type,
            name=name, tags=tags,
        )

    launcher_cls: T.Union[T.Type[AsyncLauncher], T.Type[SyncLauncher]]
    if async_mode:
        launcher_cls = AsyncLauncher
    else:
        launcher_cls = SyncLauncher

    return launcher_cls(
        func, engine, job_type,
        name, description, tags, **job_attrs,
    )



import asyncio
import functools

from loky.process_executor import ProcessPoolExecutor

from .base import Job


class ProcessJob(Job):
    """Job that runs in a process."""""

    def has_resource(self) -> bool:
        """Check if the job has enough resource to run."""
        if self.engine is None:
            return False
        else:
            return (
                super().has_resource() and
                (self.engine.resource.n_process > 0)
            )

    def consume_resource(self) -> bool:
        """Consume resource for the job."""
        if self.engine is None:
            return False
        else:
            self.engine.resource.n_process -= 1
            return (
                super().consume_resource() and
                True
            )

    def release_resource(self) -> bool:
        """Release resource for the job."""
        if self.engine is None:
            return False
        else:
            self.engine.resource.n_process += 1
            return (
                super().release_resource() and
                True
            )

    async def run(self):
        """Run job in process pool."""
        self._executor = ProcessPoolExecutor(1)
        loop = asyncio.get_running_loop()
        func = functools.partial(self.func, **self.kwargs)
        fut = loop.run_in_executor(self._executor, func, *self.args)
        result = await fut
        return result

    async def cancel(self):
        """Cancel job."""
        if self.status == "running":
            self._executor.shutdown(wait=True, kill_workers=True)
        await super().cancel()

    def clear_context(self):
        """Clear context."""
        self._executor.shutdown(wait=True, kill_workers=True)
        self._executor = None



import sys
import typing as T
from functools import update_wrapper
from pathlib import Path
import traceback


class Tee(object):
    def __init__(
            self, out_file: T.TextIO,
            stream: T.Literal['stdout', 'stderr'] = 'stdout'):
        self.stream_type = stream
        self.file = out_file
        self.stream: T.TextIO = getattr(sys, stream)

    def write(self, data):
        self.stream.write(data)
        self.file.write(data)

    def __enter__(self):
        setattr(sys, self.stream_type, self)

    def __exit__(self, _type, _value, _traceback):
        setattr(sys, self.stream_type, self.stream)


class CaptureOut(object):
    def __init__(
            self, func: T.Callable,
            stdout_file: Path, stderr_file: Path,
            capture_traceback: bool = True):
        update_wrapper(func, self)
        self.func = func
        self.stdout_file = stdout_file
        self.stderr_file = stderr_file
        self.capture_traceback = capture_traceback

    def __call__(self, *args, **kwargs) -> T.Any:
        with open(self.stdout_file, 'w') as fo, \
             open(self.stderr_file, 'w') as fe:
            with Tee(fo, 'stdout'), Tee(fe, 'stderr'):
                try:
                    res = self.func(*args, **kwargs)
                except Exception as e:
                    if self.capture_traceback:
                        traceback.print_exc(file=fe)
                    raise e
        return res



from .core import (
    launcher, SyncLauncher,
    AsyncLauncher, LauncherBase,
    set_default_engine, get_default_engine
)


__all__ = [
    'launcher', 'SyncLauncher',
    'AsyncLauncher', 'LauncherBase',
    'set_default_engine', 'get_default_engine'
]



import functools

from dask.distributed import Client, LocalCluster

from .base import Job
from ..utils import PortManager


def get_default_client() -> Client:
    free_port = PortManager.find_free_port()
    cluster = LocalCluster(
        dashboard_address=f":{free_port}",
        asynchronous=True,
    )
    return Client(
        cluster,
        asynchronous=True,
    )


class DaskJob(Job):
    """Job that runs with Dask."""

    def has_resource(self) -> bool:
        """Check if the job has enough resource to run."""
        if self.engine is None:
            return False
        else:
            return (
                super().has_resource() and
                (self.engine.resource.n_dask > 0)
            )

    def consume_resource(self) -> bool:
        """Consume resource for the job."""
        if self.engine is None:
            return False
        else:
            self.engine.resource.n_dask -= 1
            return (
                super().consume_resource() and
                True
            )

    def release_resource(self) -> bool:
        """Release resource for the job."""
        if self.engine is None:
            return False
        else:
            self.engine.resource.n_dask += 1
            return (
                super().release_resource() and
                True
            )

    async def run(self):
        """Run job with Dask."""
        client = self.engine.dask_client
        func = functools.partial(self.func, **self.kwargs)
        fut = client.submit(func, *self.args)
        self._executor = fut
        result = await fut
        return result

    async def cancel(self):
        """Cancel job."""""
        if self.status == "running":
            await self._executor.cancel()
        await super().cancel()

    def clear_context(self):
        """Clear context."""
        self._executor = None



import asyncio
import typing as T
from datetime import datetime
from pathlib import Path
from copy import copy
import itertools

import cloudpickle

from .utils import (
    JobStatusAttr, JobEmitError, InvalidStateError, JobStatusType,
    ExecutorError,
)
from .condition import Condition, AfterOthers, AllSatisfied
from ..middle.capture import CaptureOut
from ..middle.dir import ChDir
from ..log import logger
from ..base import ExecutorObj
from ..utils import get_callable_name


if T.TYPE_CHECKING:
    from ..core import Engine


class StopResolve(Exception):
    pass


class JobFuture():
    def __init__(self, job_id: str) -> None:
        """A future object for job."""
        self.job_id = job_id
        self._result: T.Optional[T.Any] = None
        self._exception: T.Optional[Exception] = None
        self.done_callbacks: T.List[T.Callable] = []
        self.error_callbacks: T.List[T.Callable] = []

    def result(self) -> T.Any:
        return self._result

    def set_result(self, result: T.Any) -> None:
        self._result = result

    def exception(self) -> T.Optional[Exception]:
        return self._exception

    def set_exception(self, exception: Exception) -> None:
        self._exception = exception

    def add_done_callback(self, callback: T.Callable) -> None:
        self.done_callbacks.append(callback)

    def add_error_callback(self, callback: T.Callable) -> None:
        self.error_callbacks.append(callback)


class Job(ExecutorObj):

    status = JobStatusAttr()
    repr_attrs: T.List[T.Tuple[str, T.Callable[["Job"], T.Any]]] = [
        ("status", lambda self: self.status),
        ("id", lambda self: self.id),
        ("func", lambda self: get_callable_name(self.func)),
        ("condition", lambda self: self.condition),
        ("retry_remain", lambda self: self.retry_remain),
    ]

    func: T.Callable
    condition: T.Optional[Condition]
    retry_remain: int

    def __init__(
            self,
            func: T.Callable,
            args: T.Optional[tuple] = None,
            kwargs: T.Optional[dict] = None,
            callback: T.Optional[T.Callable[[T.Any], None]] = None,
            error_callback: T.Optional[T.Callable[[Exception], None]] = None,
            retries: int = 0,
            retry_time_delta: float = 0.0,
            name: T.Optional[str] = None,
            condition: T.Optional[Condition] = None,
            wait_time_delta: float = 0.01,
            redirect_out_err: bool = False,
            change_dir: bool = False,
            **attrs
            ) -> None:
        """Base class for job.

        Args:
            func: The function to be executed.
            args: The positional arguments to be passed to the function.
            kwargs: The keyword arguments to be passed to the function.
            callback: The callback function to be called when the job is done.
            error_callback: The callback function to be called
                when the job is failed.
            retries: The number of retries when the job is failed.
            retry_time_delta: The time delta between retries.
            name: The name of the job.
            condition: The condition of the job.
            wait_time_delta: The time delta between each
                check of the condition.
            redirect_out_err: Whether to redirect the stdout
                and stderr to the log.
            change_dir: Whether to change the working directory
                to the log directory.
            **attrs: The attributes of the job.
        """
        super().__init__()
        self.future = JobFuture(self.id)
        self.func = func
        self.args = args or tuple()
        self.kwargs = kwargs or {}
        if callback is not None:
            self.future.add_done_callback(callback)
        if error_callback is not None:
            self.future.add_error_callback(error_callback)
        self.retries = retries
        self.retry_remain = retries
        self.retry_time_delta = retry_time_delta
        self.engine: T.Optional["Engine"] = None
        self._status: str = "created"
        self.name = name or func.__name__
        self.attrs = attrs
        self.task: T.Optional[asyncio.Task] = None
        self.condition = condition
        self.wait_time_delta = wait_time_delta
        self.redirect_out_err = redirect_out_err
        self.change_dir = change_dir
        self.created_time: datetime = datetime.now()
        self.submit_time: T.Optional[datetime] = None
        self.stoped_time: T.Optional[datetime] = None
        self._executor: T.Any = None
        self.dep_job_ids: T.List[str] = []

    def __repr__(self) -> str:
        attrs = []
        for attr_name, attr_func in self.repr_attrs:
            attr_value = attr_func(self)
            if bool(attr_value) is False:
                continue
            attrs.append(f"{attr_name}={attr_value}")
        attr_str = " ".join(attrs)
        return f"<{self.__class__.__name__} {attr_str}>"

    def __str__(self) -> str:
        return repr(self)

    def has_resource(self) -> bool:
        """Check if the job has resource to run."""
        if self.engine is None:
            return False
        else:
            return self.engine.resource.n_job > 0

    def consume_resource(self) -> bool:
        """Consume the resource of the job."""
        if self.engine is None:
            return False
        else:
            self.engine.resource.n_job -= 1
            return True

    def release_resource(self) -> bool:
        """Release the resource of the job."""
        if self.engine is None:
            return False
        else:
            self.engine.resource.n_job += 1
            return True

    def resolve_dependencies(self) -> None:
        """Resolve args and kwargs
        and auto specify the condition."""
        dep_jobs_ids: T.List[str] = []
        args = itertools.chain(self.args, self.kwargs.values())
        for arg in args:
            if isinstance(arg, JobFuture):
                dep_jobs_ids.append(arg.job_id)
        if len(dep_jobs_ids) > 0:
            after_others = AfterOthers(dep_jobs_ids)
            if self.condition is None:
                self.condition = after_others
            else:
                self.condition = AllSatisfied([self.condition, after_others])
        self.dep_job_ids = dep_jobs_ids

    async def _resolve_arg(self, arg: T.Union[JobFuture, T.Any]) -> T.Any:
        if isinstance(arg, JobFuture):
            assert self.engine is not None
            job = self.engine.jobs.get_job_by_id(arg.job_id)
            if job.status == "done":
                return job.result()
            elif job.status == "failed":
                msg = f"Job {self} cancelled because " \
                      f"of upstream job {job} failed."
                logger.warning(msg)
                await self.cancel()
                raise StopResolve(msg)
            elif job.status == "cancelled":
                msg = f"Job {self} cancelled because " \
                      f"of upstream job {job} cancelled."
                logger.warning(msg)
                await self.cancel()
                raise StopResolve(msg)
            else:  # pragma: no cover
                raise ExecutorError("Unreachable code.")
        else:
            return arg

    async def resolve_args(self):
        """Resolve args and kwargs."""
        if len(self.dep_job_ids) > 0:
            args = []
            kwargs = {}
            for arg in self.args:
                resolved = await self._resolve_arg(arg)
                args.append(resolved)
            for key, value in self.kwargs.items():
                resolved = await self._resolve_arg(value)
                kwargs[key] = resolved
            self.args = tuple(args)
            self.kwargs = kwargs

    def runnable(self) -> bool:
        """Check if the job is runnable.
        Job is runnable if:
        1. engine is not None.
        2. condition is None and condition is satisfied.
        3. has resource."""
        if self.engine is None:
            return False
        if self.condition is not None:
            return self.condition.satisfy(self.engine) and self.has_resource()
        else:
            return self.has_resource()

    async def emit(self) -> asyncio.Task:
        """Emit the job to the engine."""
        logger.info(f"Emit job {self}, watting for run.")
        if self.status != 'pending':
            raise JobEmitError(
                f"{self} is not in valid status(pending)")
        self.resolve_dependencies()
        self.submit_time = datetime.now()
        loop = asyncio.get_running_loop()
        task = loop.create_task(self.wait_and_run())
        self.task = task
        return task

    def process_func(self):
        """Process(decorate) the target func, before run.
        For example, let the func
        change the dir, redirect the stdout and stderr
        before the actual run."""
        cache_dir = self.cache_dir.resolve()
        if self.redirect_out_err and (not isinstance(self.func, CaptureOut)):
            path_stdout = cache_dir / 'stdout.txt'
            path_stderr = cache_dir / 'stderr.txt'
            self.func = CaptureOut(self.func, path_stdout, path_stderr)
        if self.change_dir:
            self.func = ChDir(self.func, cache_dir)

    async def wait_and_run(self):
        """Wait for the condition satisfied and run the job."""
        while True:
            if self.runnable() and self.consume_resource():
                logger.info(f"Start run job {self}")
                self.process_func()
                try:
                    await self.resolve_args()
                except StopResolve:
                    break
                self.status = "running"
                try:
                    res = await self.run()
                    await self.on_done(res)
                    return res
                except Exception as e:
                    await self.on_failed(e)
                    break
            else:
                await asyncio.sleep(self.wait_time_delta)

    async def run(self):
        """Run the job."""
        pass

    async def rerun(self, check_status: bool = True):
        """Rerun the job."""
        _valid_status = ("cancelled", "done", "failed")
        if check_status and (self.status not in _valid_status):
            raise JobEmitError(
                f"{self} is not in valid status({_valid_status})")
        logger.info(f"Rerun job {self}")
        self.status = "pending"
        await self.emit()

    def _on_finish(self, new_status: JobStatusType = "done"):
        self.status = new_status
        self.release_resource()

    async def on_done(self, res):
        """Callback when the job is done."""
        logger.info(f"Job {self} done.")
        self.future.set_result(res)
        for callback in self.future.done_callbacks:
            callback(res)
        self._on_finish("done")

    async def on_failed(self, e: Exception):
        """Callback when the job is failed."""
        logger.error(f"Job {self} failed: {repr(e)}")
        assert self.engine is not None
        if self.engine.print_traceback:
            logger.exception(e)
        self.future.set_exception(e)
        for err_callback in self.future.error_callbacks:
            err_callback(e)
        if self.retry_remain > 0:
            self._on_finish("pending")
            self.retry_remain -= 1
            await asyncio.sleep(self.retry_time_delta)
            await self.rerun(check_status=False)
        else:
            self._on_finish("failed")

    async def cancel(self):
        """Cancel the job."""
        logger.info(f"Cancel job {self}.")
        self.task.cancel()
        if self.status == "running":
            try:
                self.clear_context()
            except Exception as e:  # pragma: no cover
                print(str(e))
            finally:
                self._on_finish("cancelled")
        elif self.status == "pending":
            self.status = "cancelled"

    def clear_context(self):
        """Clear the context of the job."""
        pass

    def result(self) -> T.Any:
        """Get the result of the job."""
        if self.status != "done":
            raise InvalidStateError(f"{self} is not done.")
        return self.future.result()

    def exception(self):
        """Get the exception of the job."""
        return self.future.exception()

    def serialization(self) -> bytes:
        """Serialize the job to bytes."""
        job = copy(self)
        job.task = None
        job.engine = None
        job._executor = None
        bytes_ = cloudpickle.dumps(job)
        return bytes_

    @staticmethod
    def deserialization(bytes_: bytes) -> "Job":
        """Deserialize the job from bytes."""
        job: "Job" = cloudpickle.loads(bytes_)
        return job

    async def join(self, timeout: T.Optional[float] = None):
        """Wait for the job done."""
        if self.task is None:
            raise InvalidStateError(f"{self} is not emitted.")
        await asyncio.wait([self.task], timeout=timeout)

    @property
    def cache_dir(self) -> T.Optional[Path]:
        """Get the cache dir of the job."""
        if self.engine is None:
            return None
        parent = self.engine.cache_dir
        path = parent / self.id
        path.mkdir(parents=True, exist_ok=True)
        return path



from setuptools import setup, find_namespace_packages
import re


classifiers = [
    "Development Status :: 3 - Alpha",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
]


keywords = [
    'Job Management',
]


URL = "https://github.com/Nanguage/executor-engine"


def get_version():
    with open("executor/engine/__init__.py") as f:
        for line in f.readlines():
            m = re.match("__version__ = '([^']+)'", line)
            if m:
                return m.group(1)
        raise IOError("Version information can not found.")


def get_long_description():
    return f"See {URL}"


def get_requirements_from_file(filename):
    requirements = []
    with open(filename) as f:
        for line in f.readlines():
            line = line.strip()
            if len(line) == 0:
                continue
            if line and not line.startswith('#'):
                requirements.append(line)
    return requirements


def get_install_requires():
    return get_requirements_from_file('requirements.txt')


def get_doc_requires():
    return get_requirements_from_file('docs/requirements.txt')


requires_test = ['pytest', 'pytest-cov', 'flake8', 'mypy']
packages_for_dev = ["pip", "setuptools", "wheel", "twine", "ipdb"]

requires_dev = packages_for_dev + requires_test + get_doc_requires()

requires_dask = ['dask', 'distributed', 'nest_asyncio']


setup(
    name='executor-engine',
    author='Weize Xu',
    author_email='vet.xwz@gmail.com',
    version=get_version(),
    license='MIT',
    description='Package for manage job executions.',
    long_description=get_long_description(),
    keywords=keywords,
    url=URL,
    packages=find_namespace_packages(include=['executor.*']),
    include_package_data=True,
    zip_safe=False,
    classifiers=classifiers,
    install_requires=get_install_requires(),
    extras_require={
        'dev': requires_dev + requires_dask,
        'dask': requires_dask,
        'all': requires_dask,
    },
    python_requires='>=3.8, <4',
)
